## @file CommonBuildEntry.py
# This module contains code that is shared between all the entry points for PlatformBUild
# scripts.
#
##
# Copyright (c) 2017, Microsoft Corporation
#
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##
import os
import sys
import re
import logging
import subprocess
from datetime import datetime
from MuEnvironment import SelfDescribingEnvironment
from MuEnvironment import PluginManager
import pkg_resources

# Simplified Comparison Function borrowed from StackOverflow...
# https://stackoverflow.com/questions/1714027/version-number-comparison
# With Python 3.0 help from:
# https://docs.python.org/3.0/whatsnew/3.0.html#ordering-comparisons
def version_compare(version1, version2):
    def normalize(v):
        return [int(x) for x in re.sub(r'(\.0+)*$','', v).split(".")]
    (a, b) = (normalize(version1), normalize(version2))
    return (a > b) - (a < b)


#
# Shell Command with Output Helper
# This helper will attempt to run a shell command and return
# the stdout results.
# Also abstracts the bytes-vs-strings ambiguity in Python 2, 3.
# Will raise an error if return code is non-zero.
def cmd_with_output(cmd_string, cwd):
    c = subprocess.Popen(cmd_string, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=cwd, shell=True)
    c.wait()

    # Get the data.
    cmd_result = c.stdout.read()
    # PYTHON COMPAT HACK BEGIN
    if type(cmd_result) is bytes:
      cmd_result = cmd_result.decode()
    # PYTHON COMPAT HACK END

    # Check for errors.
    if c.returncode != 0:
        raise RuntimeError(cmd_result)

    return cmd_result


#
# minimum_env_init() will attempt to follow all of the steps
# necessary to get this process off the ground.
def minimum_env_init(my_workspace_path, my_project_scope):
    # TODO: Check the Git version against minimums.

    # Check the Python version against minimums.
    cur_py = "%d.%d.%d" % sys.version_info[:3]
    soft_min_py = "3.7"
    hard_min_py = "3.6"
    if version_compare(hard_min_py, cur_py) > 0:
        raise RuntimeError("Please upgrade Python! Current version is %s. Minimum is %s." % (cur_py, hard_min_py))
    if version_compare(soft_min_py, cur_py) > 0:
        logging.critical("Please upgrade Python! Current version is %s. Recommended minimum is %s." % (cur_py, soft_min_py))

    # Initialized the build environment.
    return SelfDescribingEnvironment.BootstrapEnvironment(my_workspace_path, my_project_scope)


#
# configure_base_logging() sets up only the logging that will
# be used by all commands. (ie. doesn't configure any logging files)
def configure_base_logging(mode="standard"):
    # Initialize logging.
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG)

    # Adjust console mode depending on mode.
    console = logging.StreamHandler()
    if mode == "vs":
        console.setLevel(logging.DEBUG)
        console.setFormatter(logging.Formatter("%(message)s"))
    elif mode == 'verbose':
        console.setLevel(logging.DEBUG)
        console.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
    elif mode == 'simple':
        console.setLevel(logging.INFO)
        console.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
    else:
        console.setLevel(logging.CRITICAL)
        console.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))

    # Add the console as a logger, now that it's configured.
    logger.addHandler(console)

#
# setup_process() automates all of the processes that should be unique
# to each platform build. It will attempt to set up the repos and
# anything else that's important.
def setup_process(my_workspace_path, my_project_scope, my_required_repos, force_it=False):
    # Grab the remaining Git repos.
    if my_required_repos:
        #### Git Repos: STEP 1 --------------------------------------
        # Make sure that the repos are all synced.
        print("## Syncing Git repositories: %s..." % ", ".join(my_required_repos))
        cmd_parts = ("git", "submodule", "sync")
        cmd_parts += my_required_repos
        cmd_string = " ".join(cmd_parts)
        try:
            c = subprocess.Popen(cmd_string, stderr=subprocess.STDOUT, cwd=my_workspace_path, shell=True)
            c.wait()
            if c.returncode != 0:
                raise RuntimeError(c.stdout.read())
        except:
            print("FAILED!\n")
            print("Failed to sync required repositories!\n")
            raise
        print("Done.\n")

        #### Git Repos: STEP 2 --------------------------------------
        # Iterate through all repos and see whether they should be fetched.
        for required_repo in my_required_repos:
            print("## Checking Git repository: %s..." % required_repo)

            #### Git Repos: STEP 2a ---------------------------------
            # Need to determine whether to skip this repo.
            repo_path = os.path.join(my_workspace_path, required_repo)
            skip_repo = False
            # If the repo exists (and we're not forcing things) make
            # sure that it's not in a "dirty" state.
            if os.path.exists(repo_path) and not force_it:
                git_data = None
                cmd_parts = ("git", "diff", required_repo)
                cmd_string = " ".join(cmd_parts)
                try:
                    git_data = cmd_with_output(cmd_string, my_workspace_path)
                except:
                    # We don't actually care. We'll handle repo errors later.
                    # We only care if we can get the data.
                    pass

                # If anything was returned, we should skip processing the repo.
                # It is either on a different commit or it has local changes.
                if git_data:
                    print("-- NOTE: Repo currently exists and appears to have local changes!")
                    print("-- Skipping fetch!")
                    skip_repo = True

            #### Git Repos: STEP 2b ---------------------------------
            # If we're not skipping, grab it.
            if not skip_repo or force_it:
                print("## Fetching repo.")
                cmd_parts = ("git", "submodule", "update", "--init", "--recursive", "--progress", required_repo)
                cmd_string = " ".join(cmd_parts)
                try:
                    c = subprocess.Popen(cmd_string, stderr=subprocess.STDOUT, cwd=my_workspace_path, shell=True)
                    c.wait()
                    if c.returncode != 0:
                        raise RuntimeError(c.stdout.read())
                except:
                    print("FAILED!\n")
                    print("Failed to fetch required repositories!\n")
                    raise
            print("Done.\n")

    # Now that we should have all of the required code,
    # we're ready to build the environment and fetch the
    # dependencies for this project.
    print("## Fetching all external dependencies...")
    (build_env, shell_env) = minimum_env_init(my_workspace_path, my_project_scope)
    SelfDescribingEnvironment.UpdateDependencies(my_workspace_path, my_project_scope)
    print("Done.\n")

    # TODO: Install any certs any other things that might be required.


def update_process(my_workspace_path, my_project_scope):
    # Get the environment set up.
    logging.info("## Parsing environment...")
    (build_env, shell_env) = minimum_env_init(my_workspace_path, my_project_scope)
    logging.info("Done.\n")

    # Update the environment.
    logging.info("## Updating environment...")
    SelfDescribingEnvironment.UpdateDependencies(my_workspace_path, my_project_scope)
    logging.info("Done.\n")


def build_process(my_workspace_path, my_project_scope, my_module_pkg_paths):
    #
    # Initialize file-based logging.
    #
    logfile = os.path.join(my_workspace_path, "Build", "BUILDLOG.TXT")
    if not os.path.isdir(os.path.dirname(logfile)):
        os.makedirs(os.path.dirname(logfile))

    filelogger = logging.FileHandler(filename=(logfile), mode='w')
    filelogger.setLevel(logging.DEBUG)
    filelogger.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
    logging.getLogger('').addHandler(filelogger)

    logging.info("Log Started: " + datetime.strftime(datetime.now(), "%A, %B %d, %Y %I:%M%p" ))
    logging.info("Running Python version: " + str(sys.version_info))
    logging.info("Running mu_python_library version: " + pkg_resources.get_distribution("mu_python_library").version)
    logging.info("Running mu_environment version: " + pkg_resources.get_distribution("mu_environment").version)

    #
    # Next, get the environment set up.
    #
    try:
        (build_env, shell_env) = minimum_env_init(my_workspace_path, my_project_scope)
        if not SelfDescribingEnvironment.VerifyEnvironment(my_workspace_path, my_project_scope):
            raise RuntimeError("Validation failed.")
    except:
        raise RuntimeError("Environment is not in a state to build! Please run '--UPDATE'.")

    #Load plugins
    pluginManager = PluginManager.PluginManager()
    pluginManager.SetListOfEnvironmentDescriptors(build_env.plugins)
    helper = PluginManager.HelperFunctions()
    helper.LoadFromPluginManager(pluginManager)
    pluginList = pluginManager.GetPluginsOfClass(PluginManager.IMuBuildPlugin)

    # NOTE: This implicitly assumes that the PlatformBuild script path is in PYTHONPATH.
    from PlatformBuildWorker import PlatformBuilder

    #
    # Now we can actually kick off a build.
    #
    PB = PlatformBuilder(my_workspace_path, my_module_pkg_paths, pluginManager, helper, sys.argv)
    retcode = PB.Go()

    if(retcode != 0):
        logging.critical("Error")
        logging.critical("Log file at " + logfile)
    else:
        logging.critical("Success")

    #get all vars needed as we can't do any logging after shutdown otherwise our log is cleared.  
    #Log viewer
    ep = PB.env.GetValue("LaunchBuildLogProgram")
    LogOnSuccess = PB.env.GetValue("LaunchLogOnSuccess")
    LogOnError = PB.env.GetValue("LaunchLogOnError")
    
    #end logging
    logging.shutdown()
    #no more logging

    if(ep != None):
        cmd = ep + " " + logfile

    #
    # Conditionally launch the shell to show build log
    #
    #
    if( ((retcode != 0) and (LogOnError.upper() == "TRUE")) or (LogOnSuccess.upper() == "TRUE")):
        subprocess.Popen(cmd, shell=True)
        
    sys.exit(retcode)


def build_entry(my_script_path, my_workspace_path, my_required_repos, my_project_scope, my_module_pkgs, my_module_pkg_paths):
    logging_mode = "standard"
    script_process = "build"
    force_process = False

    # Check for some well-known parameters.
    for arg in sys.argv:
        if "--ANDUPDATE" == arg.upper():
            script_process = "and_update"
        if "--FORCE" == arg.upper():
            force_process = True
        if "--SETUP" == arg.upper():
            logging_mode = "simple"
            script_process = "setup"
        if "--UPDATE" == arg.upper():
            logging_mode = "simple"
            script_process = "update"
        if "--VERBOSE" == arg.upper():
            logging_mode = "verbose"
        if "--VSMODE" == arg.upper():
            logging_mode = "vs"

    # TODO: Scrub the parameters so they're not passed on to the next script.

    # Turn on logging for the remainder of the process.
    configure_base_logging(logging_mode)

    # Execute the requested process.
    if script_process == "setup":
        setup_process(my_workspace_path, my_project_scope, my_required_repos, force_it=force_process)
    elif script_process == "update":
        update_process(my_workspace_path, my_project_scope)
    else:
        if script_process == "and_update":
            update_process(my_workspace_path, my_project_scope)
        build_process(my_workspace_path, my_project_scope, my_module_pkg_paths)



## @file UefiBuild.py
# This module contains code that supports the Tianocore Edk2 build system
# This class is designed to be subclassed by a platform to allow
# more extensive and custom behavior. 
#
# TODO: more docs
#
##
# Copyright (c) 2018, Microsoft Corporation
#
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##


import logging
from MuEnvironment.MultipleWorkspace import *
from MuEnvironment import ConfMgmt
import subprocess
import traceback
import shutil
import threading
import time
from datetime import datetime
from datetime import timedelta
from MuEnvironment import ShellEnvironment
from PythonLibrary.Uefi.EdkII.Parsers.TargetTxtParser import *
from PythonLibrary.Uefi.EdkII.Parsers.DscParser import *
from PythonLibrary.UtilityFunctions import RunCmd

from MuEnvironment import PluginManager

class UefiBuilder(object):

    ##
    #
    # 
    # - PackagesPath - os.pathsep string containing packages path
    def __init__(self, WorkSpace, PackagesPath, PInManager, PInHelper, args, BuildConfigFile=None):
        self.env = ShellEnvironment.GetBuildVars()
        self.mws = MultipleWorkspace()
        self.mws.setWs(WorkSpace, PackagesPath)
        self.ws = WorkSpace
        self.pp = PackagesPath  #string using os.pathsep
        self.Args = args
        self.SkipBuild = False
        self.SkipPreBuild = False
        self.SkipPostBuild = False
        self.FlashImage = False
        self.ShowHelpOnly = False
        self.OutputBuildEnvBeforeBuildToFile = None
        self.Clean = False
        self.UpdateConf = False
        self.Helper = PInHelper
        self.PluginManager = PInManager
        if(BuildConfigFile != None):
            self.BuildConfig = BuildConfigFile
        else:
            self.BuildConfig = os.path.join(self.ws, "BuildConfig.conf")
        self.RunCmd = RunCmd

    def Go(self):
        try:
            self.ParseForHelp()
            if(self.ShowHelpOnly == True):
                self.ShowHelp()
                return 0

            self.Helper.DebugLogRegisteredFunctions()
            
            ret = self.SetEnv()
            if(ret != 0):
                logging.critical("SetEnv failed")
                return ret

            #clean
            if(self.Clean == True):
                logging.critical("Cleaning")
                ret = self.CleanTree()
                if(ret != 0):
                    logging.critical("Clean failed")
                    return ret
            
            #prebuild
            if(self.SkipPreBuild == True):
                logging.critical("Skipping Pre Build")
            else:
                ret = self.PreBuild()
                if(ret != 0):
                    logging.critical("Pre Build failed")
                    return ret

            # Output Build Environment to File - this is mostly for debug of build
            # issues or adding other build features using existing variables
            if(self.OutputBuildEnvBeforeBuildToFile is not None):
                logging.critical("Writing Build Env Info out to File")
                logging.debug("Found an Output Build Env File: " + self.OutputBuildEnvBeforeBuildToFile)
                self.env.PrintAll(self.OutputBuildEnvBeforeBuildToFile)

            if(self.env.GetValue("GATEDBUILD") is not None) and (self.env.GetValue("GATEDBUILD").upper() == "TRUE"):
                ShouldGatedBuildRun = self.PlatformGatedBuildShouldHappen()
                logging.debug("Platform Gated Build Should Run returned: %s" % str(ShouldGatedBuildRun))
                if( self.SkipBuild == False):
                    self.SkipBuild = not ShouldGatedBuildRun
                if(self.SkipPostBuild == False):
                    self.SkipPostBuild = not ShouldGatedBuildRun

            #build
            if(self.SkipBuild == True):
                logging.critical("Skipping Build")
            else:
                ret = self.Build()

                if(ret != 0):
                    logging.critical("Build failed")
                    return ret

            #postbuild
            if(self.SkipPostBuild == True):
                logging.critical("Skipping Post Build")
            else:
                ret = self.PostBuild()
                if(ret != 0):
                    logging.critical("Post Build failed")
                    return ret

            #flash
            if(self.FlashImage == True):
                logging.critical("Flashing Image")
                ret = self.FlashRomImage()
                if(ret != 0):
                    logging.critical("Flash Image failed")
                    return ret

        except:            
            logging.critical("Build Process Exception")
            logging.debug(traceback.format_exc())
            return -1

        return 0
                    

    def CleanTree(self, RemoveConfTemplateFilesToo=False):
        ret = 0
        #loop thru each build target set.  
        logging.critical("Cleaning All Output for Build")

        d = self.env.GetValue("BUILD_OUTPUT_BASE")
        if(os.path.isdir(d)):
            logging.debug("Removing [%s]", d)
            # if the folder is opened in Explorer do not fail the entire Rebuild
            try:
                shutil.rmtree(d)
            except WindowsError as wex:
                logging.debug(wex)

        else:
            logging.debug("Directory [%s] already clean" % d)

        #delete the conf .dbcache       
        #this needs to be removed in case build flags changed
        d = os.path.join(self.ws, "Conf", ".cache")
        if(os.path.isdir(d)):
            shutil.rmtree(d)
            logging.debug("Removing [%s]" % d)

        if(RemoveConfTemplateFilesToo):
            for a in ["target.txt", "build_rule.txt", "tools_def.txt"]:
                d = os.path.join(self.ws, "Conf", a)
                if(os.path.isfile(d)):
                    os.remove(d)
                    logging.debug("Removing [%s]" % d)

        return ret
                

    #
    # Build step
    #
    def Build(self): 
        BuildType = self.env.GetValue("TARGET")
        logging.critical("Running Build %s" % BuildType)
            
        cmd = "build "
        #set target, arch toolchain
        cmd = cmd + " -p " + self.env.GetValue("ACTIVE_PLATFORM")
        cmd = cmd + " -b " + BuildType
        cmd = cmd + " -t " + self.env.GetValue("TOOL_CHAIN_TAG")

        #Set the arch flags.  Multiple are split by space
        rt = self.env.GetValue("TARGET_ARCH").split(" ")
        for t in rt:
            cmd = cmd + " -a " + t
            
        #get the report options and setup the build command
        if(self.env.GetValue("BUILDREPORTING") == "TRUE"):
            cmd = cmd + " -y " + self.env.GetValue("BUILDREPORT_FILE")
            rt = self.env.GetValue("BUILDREPORT_TYPES").split(" ")
            for t in rt:
                cmd = cmd + " -Y " + t

        #add special processing to handle building a single module
        mod = self.env.GetValue("BUILDMODULE")
        if( mod != None and len(mod.strip()) > 0):
            cmd = cmd + " -m " + mod
            logging.critical("Single Module Build: " + mod )
            self.SkipPostBuild = True
            self.FlashImage = False

        #attach the generic build vars
        buildvars = self.env.GetAllBuildKeyValues(BuildType)
        for key,value in buildvars.items():
            cmd = cmd + " -D " + key + "=" + value
        ret = self.RunCmd(cmd)
        if(ret != 0):
            return ret

        return 0

    def PreBuild(self):
        logging.critical("Running Pre Build")
        #
        # Run the plaform pre-build steps.
        #
        ret = self.PlatformPreBuild()

        if(ret != 0):
            logging.critical("PlatformPreBuild failed %d" % ret)
            return ret
        #
        # run all loaded UefiBuild Plugins
        #
        for Descriptor in self.PluginManager.GetPluginsOfClass(PluginManager.IUefiBuildPlugin):
            
            rc = Descriptor.Obj.do_pre_build(self)
            if(rc != 0):
                if(rc is None):
                    logging.error("Plugin Failed: %s returned NoneType" % Descriptor.Name)
                    ret = -1
                else:
                    logging.error("Plugin Failed: %s returned %d" % (Descriptor.Name, rc))
                    ret = rc
            else:
                logging.debug("Plugin Success: %s" % Descriptor.Name)
        return ret

    def PostBuild(self):
        logging.critical("Running Post Build")
        #
        # Run the platform post-build steps.
        #
        ret = self.PlatformPostBuild()

        if(ret != 0):
            logging.critical("PlatformPostBuild failed %d" % ret)
            return ret

        #
        # run all loaded UefiBuild Plugins
        #
        for Descriptor in self.PluginManager.GetPluginsOfClass(PluginManager.IUefiBuildPlugin):
            
            rc = Descriptor.Obj.do_post_build(self)
            if(rc != 0):
                if(rc is None):
                    logging.error("Plugin Failed: %s returned NoneType" % Descriptor.Name)
                    ret = -1
                else:
                    logging.error("Plugin Failed: %s returned %d" % (Descriptor.Name, rc))
                    ret = rc
            else:
                logging.debug("Plugin Success: %s" % Descriptor.Name)

        return ret
    

    def SetEnv(self):
        logging.critical("Setting up the Environment")
        ShellEnvironment.GetEnvironment().set_shell_var("WORKSPACE", self.ws)
        ShellEnvironment.GetBuildVars().SetValue("WORKSPACE", self.ws, "Set in SetEnv")

        #1. Process command line parameters
        ret = self.ParseInputArgs()
        if (ret != 0):
            logging.critical("Parse Input Args Failed")
            return ret

        #look for config file and parse each line
        ret = self.ParseCustomConfigFile()
        if(ret != 0):
            logging.critical("Parse custom config file failed")
            return ret

        if(self.pp is not None):
            ShellEnvironment.GetEnvironment().set_shell_var("PACKAGES_PATH", self.pp)
            ShellEnvironment.GetBuildVars().SetValue("PACKAGES_PATH", self.pp, "Set in SetEnv")

        #process platform parameters defined in platform build file
        ret = self.SetPlatformEnv()
        if(ret != 0):
            logging.critical("Set Platform Env failed")
            return ret

        #set some basic defaults 
        self.SetBasicDefaults()

        #Handle all the template files for workspace/conf/ Allow override 
        TemplatesForConf = self.env.GetValue("CONF_TEMPLATE_DIR")
        if(TemplatesForConf is not None):
            TemplatesForConf = self.mws.join(self.ws, TemplatesForConf)
            logging.debug("Platform defined override for Template Conf Files: %s", TemplatesForConf)
        e = ConfMgmt.ConfMgmt(self.UpdateConf, TemplatesForConf)

        #parse target file
        ret = self.ParseTargetFile()
        if(ret != 0):
            logging.critical("ParseTargetFile failed")
            return ret

        ret = e.ToolsDefConfigure()
        if(ret != 0):
            logging.critical("ParseTargetFile failed")
            return ret

        #parse DSC file
        ret = self.ParseDscFile()
        if(ret != 0):
            logging.critical("ParseDscFile failed")
            return ret

        #parse FDF file
        ret = self.ParseFdfFile()
        if(ret != 0):
            logging.critical("ParseFdfFile failed")
            return ret
        
        #set build output base envs for all builds
        self.env.SetValue("BUILD_OUT_TEMP", os.path.join(self.ws, self.env.GetValue("OUTPUT_DIRECTORY")), "Computed in SetEnv")
        
        target = self.env.GetValue("TARGET")
        self.env.SetValue("BUILD_OUTPUT_BASE", os.path.join(self.env.GetValue("BUILD_OUT_TEMP"), target + "_" + self.env.GetValue("TOOL_CHAIN_TAG")), "Computed in SetEnv")

        #We have our build target now.  Give platform build one more chance for target specific settings.
        ret = self.SetPlatformEnvAfterTarget()
        if(ret != 0):
            logging.critical("SetPlatformEnvAfterTarget failed")
            return ret

        #set the build report file
        self.env.SetValue("BUILDREPORT_FILE", os.path.join(self.env.GetValue("BUILD_OUTPUT_BASE"), "BUILD_REPORT.TXT"), True)
        
        #set environment variables for the build process
        os.environ["EFI_SOURCE"] = self.ws
        
        return 0

    def ParseCustomConfigFile(self):
        fp = self.BuildConfig
        if(os.path.isfile(fp)):
            f = open(fp, "r")
            for l in f.readlines():
                line = l.strip().partition("#")[0]
                if(len(line) < 1):
                    continue

                (key,sep,value) = line.partition('=')
                if(len(key) < 1):
                    logging.error("Key invalid in Custom Config file: " + l)
                    continue

                if(len(value) < 1):
                    logging.error("Value invalid in Custom Config File: " + l)
                    continue
                
                self.env.SetValue(key.strip().upper(), value.strip(), "From MyBuildConfig.conf")

            f.close()
            
        return 0
    
    def FlashRomImage(self):
        return self.PlatformFlashImage()
        
        

    #-----------------------------------------------------------------------
    # Methods that will be overridden by child class
    #-----------------------------------------------------------------------
    
    @classmethod
    def PlatformPreBuild(self):
        return 0

    @classmethod
    def PlatformPostBuild(self):
        return 0

    @classmethod
    def SetPlatformEnv(self):
        return 0

    @classmethod
    def SetPlatformEnvAfterTarget(self):
        return 0

    @classmethod
    def PlatformBuildRom(self):
        return 0

    @classmethod
    def PlatformFlashImage(self):
        return 0

    @classmethod
    def PlatformGatedBuildShouldHappen(self):
        return True

    #------------------------------------------------------------------------
    #  HELPER FUNCTIONS
    #------------------------------------------------------------------------
    #

    #
    # Parse the TargetText file and add them as env settings.
    # set them so they can be overridden. 
    #
    def ParseTargetFile(self):
        if(os.path.isfile(self.mws.join(self.ws, "Conf", "target.txt"))):
             #parse TargetTxt File
            logging.debug("Parse Target.txt file")
            ttp = TargetTxtParser()
            ttp.ParseFile(self.mws.join(self.ws, "Conf", "target.txt"))
            for key,value in ttp.Dict.items():
                #set env as overrideable
                self.env.SetValue(key, value, "From Target.txt", True)

        else:
            logging.error("Failed to find target.txt file")
            return -1

        return 0


    #
    # Parse the Active platform DSC file.  This will get lots of variable info to
    # be used in the build.  This makes it so we don't have to define things twice
    #
    def ParseDscFile(self):
        dsc_file_path = self.mws.join(self.ws, self.env.GetValue("ACTIVE_PLATFORM"))
        if(os.path.isfile(dsc_file_path)):
             #parse DSC File
            logging.debug("Parse Active Platform DSC file: {0}".format(dsc_file_path))
            dscp = DscParser().SetBaseAbsPath(self.ws).SetPackagePaths(self.pp.split(os.pathsep)).SetInputVars(self.env.GetAllBuildKeyValues())
            dscp.ParseFile(dsc_file_path)
            for key,value in dscp.LocalVars.items():
                #set env as overrideable
                self.env.SetValue(key, value, "From Platform DSC File", True)

        else:
            logging.error("Failed to find DSC file")
            return -1

        return 0

    #
    # Parse the Active platform FDF file.  This will get lots of variable info to
    # be used in the build.  This makes it so we don't have to define things twice
    # the FDF file usually comes from the Active Platform DSC file so it needs to 
    # be parsed first.  
    #
    def ParseFdfFile(self):
        if(self.env.GetValue("FLASH_DEFINITION") is None):
            logging.debug("No flash definition set")
            return 0
        if(os.path.isfile(self.mws.join(self.ws, self.env.GetValue("FLASH_DEFINITION")))):
            #parse the FDF file- fdf files have similar syntax to DSC and therefore parser works for both.
            logging.debug("Parse Active Flash Definition (FDF) file")
            fdfp = DscParser().SetBaseAbsPath(self.ws).SetPackagePaths(self.pp.split(os.pathsep)).SetInputVars(self.env.GetAllBuildKeyValues())
            pa = self.mws.join(self.ws, self.env.GetValue("FLASH_DEFINITION"))
            fdfp.ParseFile(pa)
            for key, value in fdfp.LocalVars.items():
                self.env.SetValue(key, value, "From Platform FDF File", True)

        else:
            logging.error("Failed to find FDF file")
            return -2

        return 0

    #
    # Function used to set default values for numerous build
    # flow control variables
    #
    def SetBasicDefaults(self):
        self.env.SetValue("WORKSPACE", self.ws, "DEFAULT")
        if(self.pp is not None):
            self.env.SetValue("PACKAGES_PATH", self.pp, "DEFAULT")
        return 0

    #
    # Show the help
    #
    def ShowHelp(self):
        logging.critical("Showing help")
        print("------------------------------------------")
        print("  UefiBuild Help")
        print("------------------------------------------")
        print(" -h, --Help, -?              - Show this")
        print(" <key>=<value>               - Set an env variable for the pre/post build process")
        print(" BLD_*_<key>=<value>         - Set a build flag for all build types.  Key=value will get passed to build process")
        print(" BLD_<TARGET>_<key>=<value>  - Set a build flag for build type of <target>.  Key=value will get passed to build process for given build type")
        print(" --skipbuild                 - Skip the build process ")
        print(" --skipprebuild              - Skip prebuild process")
        print(" --skippostbuild             - Skip postbuild process")
        print(" --FlashRom                  - Flash rom after build.  Only works with single target")
        print(" --FlashOnly                 - Flash rom.  Rom must be built previously.  Only works with single target")
        print(" --UpdateConf                - Update Conf.  Builders Conf files will be replaced with latest template files")
        print(" --Clean                     - Clean.  Remove all old build artifacts and intermediate files")
        print(" --CleanOnly                 - Clean Only.  Do clean operation and don't build just exit.")

    #
    # Parse args looking for the help flag
    #
    def ParseForHelp(self):
        for a in self.Args[1:]:
            if(a.startswith("-")):
                if(a.upper() == "--HELP"):
                    self.ShowHelpOnly = True
                elif(a.upper() == "-H"):
                    self.ShowHelpOnly = True
                elif(a.upper() == "-?"):
                    self.ShowHelpOnly = True           
        
    #
    # Parse out any arguments so they can be set into the env
    #
    def ParseInputArgs(self):
        #skip first arg as its program name
        for a in self.Args[1:]:
            if(a.startswith("--")):
                if(a.upper() == "--SKIPBUILD"):
                    self.SkipBuild = True
                elif(a.upper() == "--SKIPPREBUILD"):
                    self.SkipPreBuild = True
                elif(a.upper() == "--SKIPPOSTBUILD"):
                    self.SkipPostBuild = True
                elif(a.upper() == "--FLASHONLY"):
                    self.SkipPostBuild = True
                    self.SkipBuild = True
                    self.SkipPreBuild = True
                    self.FlashImage = True
                elif(a.upper() == "--FLASHROM"):
                    self.FlashImage = True
                elif(a.upper() == "--UPDATECONF"):
                    self.UpdateConf = True
                elif(a.upper() == "--CLEAN"):
                    self.Clean = True
                elif(a.upper() == "--CLEANONLY"):
                    self.Clean = True
                    self.SkipBuild = True
                    self.SkipPreBuild = True
                    self.SkipPostBuild = True
                    self.FlashImage = False
                elif(a.upper().startswith("--BUILD_ENV_OUT_FILE")) and (a.count("=") == 1):
                    self.OutputBuildEnvBeforeBuildToFile = a.partition("=")[2]
                    
            elif(a.count("=") == 1):
                tokens = a.strip().split("=")
                self.env.SetValue(tokens[0], tokens[1], "From CmdLine")
            else:
                logging.critical("Unknown build parameter!!  Parameter: %s" % a)
                return -1
        return 0




## @file ConfMgmt.py
# Handle Edk2 Conf management
# Customized for Project Mu including dynamic support for VS2017
#
##
# Copyright (c) 2018, Microsoft Corporation
#
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##
import os
import logging
import shutil
from MuEnvironment import ShellEnvironment
import time
from PythonLibrary.UtilityFunctions import RunCmd
try: 
    from StringIO import StringIO
except ImportError:
    from io import StringIO



class ConfMgmt():
    
    def __init__(self, OverrideConf, AdditionalTemplateConfDir):
        self.Logger = logging.getLogger("ConfMgmt")
        self.env = ShellEnvironment.GetBuildVars()
        self.__PopulateConf(OverrideConf, AdditionalTemplateConfDir)

    #
    # Compare the version of the existing conf file to the template file
    #
    def __OlderVersion(self, confFile, confTemplateFile):
        conf = 0
        template = 0

        f = open(confFile, "r")
        for l in f.readlines():
            if(l.startswith("#!VERSION=")):
                conf = float(l.split("=")[1].split()[0].strip())
                logging.debug("Conf version: %s", str(conf))
                break

        f.close()
        f = open(confTemplateFile, "r")
        for l in f.readlines():
            if(l.startswith("#!VERSION=")):
                template = float(l.split("=")[1].split()[0].strip())
                logging.debug("Template Version: %s", str(template))
                break
        f.close()

        return (conf < template)

                    
        

    def __PopulateConf(self, OverrideConf, AdditionalTemplateConfDir):
        ws = self.env.GetValue("WORKSPACE")
        #Copy Conf template files to conf if not present
        target = os.path.join(ws, "Conf", "target.txt")
        buildrules = os.path.join(ws, "Conf", "build_rule.txt")
        toolsdef = os.path.join(ws, "Conf", "tools_def.txt")

        #BaseTools Template files
        target_template = os.path.join("Conf", "target.template")
        tools_def_template = os.path.join("Conf", "tools_def.template")
        build_rules_template = os.path.join("Conf", "build_rule.template")

        outfiles = [target, toolsdef, buildrules]
        tfiles = [target_template, tools_def_template, build_rules_template]

        #check if conf exists
        if( not os.path.isdir(os.path.join(ws, "Conf"))):
            os.mkdir(os.path.join(ws, "Conf"))
        
        x = 0
        while(x < len(outfiles)):
            #check if the conf file already exists
            #don't overwrite if exists.  Popup if version is older in conf
            TemplateFilePath = ""
            ConfFilePath = outfiles[x]


            #
            # Get the Override template if it exist
            #
            if(AdditionalTemplateConfDir is not None):
                fp = os.path.join(AdditionalTemplateConfDir, tfiles[x] + ".ms")
                if os.path.isfile(fp):
                    TemplateFilePath = fp

            #
            # If not found above try MS templates
            #
            if(TemplateFilePath == ""):
                fp = os.path.join(self.env.GetValue("EDK2_BASE_TOOLS_DIR"), tfiles[x] + ".ms")
                if os.path.isfile(fp):
                    TemplateFilePath = fp

            #
            # If not found above try TianoCore Template
            #
            if(TemplateFilePath == ""):
                fp = os.path.join(self.env.GetValue("EDK2_BASE_TOOLS_DIR"), tfiles[x])
                if TemplateFilePath == "" and os.path.isfile(fp):
                    TemplateFilePath = fp

            #
            # Check to see if found yet -- No more options so now we are broken
            #
            if(TemplateFilePath == ""):  
                self.Logger.critical("Failed to find Template file for %s" % outfiles[x])
                raise Exception("Template File Missing" , outfiles[x])
            else:
                self.Logger.debug("Conf file template: [%s]", TemplateFilePath)

            #Check to see if we need the template
            if(not os.path.isfile(outfiles[x])):
                #file doesn't exist.  copy template
                self.Logger.debug("%s file not found.  Creating from Template file %s" % (outfiles[x], TemplateFilePath ))
                shutil.copy2(TemplateFilePath, outfiles[x])

            elif(OverrideConf):
                self.Logger.debug("%s file replaced as requested" % outfiles[x])
                shutil.copy2(TemplateFilePath, outfiles[x])
            else:
                #Both file exists.  Do a quick version check
                if(self.__OlderVersion(outfiles[x], TemplateFilePath)):
                    #Conf dir is older.  Warn user.
                    self.Logger.critical("Conf file [%s] out-of-date.  Please update your conf files!  Sleeping 30 seconds to encourage update....", outfiles[x])
                    time.sleep(30)
                else:
                    self.Logger.debug("Conf file [%s] up-to-date", outfiles[x])
                    
            x = x + 1
        #end of while loop

    #
    # Use VsWhere tool to find visual studio tools
    #
    # return tuple of (error code, string value)
    #
    def FindWithVsWhere(self, products=None):
        cmd = "VsWhere -latest -nologo -all -property installationPath"
        if(products is not None):
            cmd += " -products " + products
        a = StringIO()
        ret = RunCmd(cmd, outstream=a)
        if(ret != 0):
            self.Logger.error("Failed in VsWhere %d to get install dir" % ret)
            a.close()
            return (ret, None)
        p1 = a.getvalue().strip()
        a.close()
        if( len(p1.strip()) > 0):
            return (0, p1)
        return (ret, None)

    def ToolsDefConfigure(self):
        Tag = self.env.GetValue("TOOL_CHAIN_TAG")
        if (Tag is not None) and (Tag.upper().startswith("VSLATEST")):
            p1 = None
            self.Logger.debug("Must find latest VS toolchain")
            for p in [None, "Microsoft.VisualStudio.Product.BuildTools", "*"]:
                (rc, path) = self.FindWithVsWhere(p)
                if rc == 0 and path is not None:
                    self.Logger.debug("Found VS instance using products = %s", p)
                    p1 = path
                    break
            if(p1 is None):
                self.Logger.critical("Failed to find valid VSLatest instance")
                return -6
            
            self.Logger.debug("VS150INSTALLPATH is %s" % p1)
            os.environ["VS150INSTALLPATH"] = p1
            #now get vc version
            p2 = os.path.join(p1, "VC", "Tools", "MSVC")
            if not os.path.isdir(p2):
                self.Logger.critical("Failed to find VC tools.  Might need to check for VS install")
                return -5
            newest = os.listdir(p2)[-1]  #get last in list
            self.Logger.debug("VS150TOOLVER is %s" % newest)
            os.environ["VS150TOOLVER"] = newest.strip()
        else:
            self.Logger.debug("Tool Chain Tag not set or not vs latest")
        return 0




## @file ShellEnvironment.py
# This module contains code that helps to manage the build environment
# including PATH, PYTHONPATH, and ENV variables.
#
##
# Copyright (c) 2017, Microsoft Corporation
#
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##
import os
import sys
import logging
from MuEnvironment import VarDict


# NOTE: It's entirely possible that this should be a singleton.
#       See here for more ideas...
#       https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
class ShellEnvironment(object):
  def __init__(self):
    super(ShellEnvironment, self).__init__()

    self.global_path = None
    self.global_pypath = None

    self.import_environment()

    self.original_path = self.global_path
    self.original_pypath = self.global_pypath

  def set_path(self, path_elements):
    logging.debug("Overriding PATH with new value.")
    self.global_path = list(path_elements)
    os.environ["PATH"] = os.pathsep.join(path_elements)
    self.export_environment()

  def set_pypath(self, path_elements):
    logging.debug("Overriding PYTHONPATH with new value.")
    self.global_pypath = list(path_elements)
    self.export_environment()

  def append_path(self, path_element):
    logging.debug("Appending PATH element '%s'." % path_element)
    if path_element not in self.global_path:
      self.global_path.append(path_element)
      self.export_environment()

  def append_pypath(self, path_element):
    logging.debug("Appending PYTHONPATH element '%s'." % path_element)
    if path_element not in self.global_pypath:
      self.global_pypath.append(path_element)
      self.export_environment()

  def insert_path(self, path_element):
    logging.debug("Inserting PATH element '%s'." % path_element)
    if path_element not in self.global_path:
      self.global_path.insert(0, path_element)
      self.export_environment()

  def insert_pypath(self, path_element):
    logging.debug("Inserting PYTHONPATH element '%s'." % path_element)
    if path_element not in self.global_path:
      self.global_pypath.insert(0, path_element)
      self.export_environment()

  def get_build_var(self, var_name):
    return GetBuildVars().GetValue(var_name)

  def set_build_var(self, var_name, var_data):
    logging.debug("Updating BUILD VAR element '%s': '%s'." % (var_name, var_data))
    GetBuildVars().SetValue(var_name, var_data, '', overridable=True)

  def get_shell_var(self, var_name):
    return os.environ.get(var_name, None)

  def set_shell_var(self, var_name, var_data):
    logging.debug("Updating SHELL VAR element '%s': '%s'." % (var_name, var_data))
    os.environ[var_name] = var_data

  def import_environment(self):
    # Record the PATH elements of the current environment.
    path = os.environ.get('PATH', "")
    # Filter removes empty elements.
    # List creates an actual list rather than a generator.
    self.global_path = list(filter(None, path.split(os.pathsep)))  # Filter removes empty strings.

    # Record the PYTHONPATH elements of the current environment.
    # When reading PYTHONPATH, try reading the live path from sys.
    self.global_pypath = sys.path

  def export_environment(self):
    os.environ["PATH"] = os.pathsep.join(self.global_path)
    os.environ["PYTHONPATH"] = os.pathsep.join(self.global_pypath)
    sys.path = self.global_pypath

  def log_environment(self):
    logging.debug("FINAL PATH:")
    logging.debug(", ".join(self.global_path))

    logging.debug("FINAL PYTHONPATH:")
    logging.debug(", ".join(self.global_pypath))


rootEnvironment = None
def GetEnvironment():
  global rootEnvironment

  if not rootEnvironment:
    rootEnvironment = ShellEnvironment()
  return rootEnvironment

  if len(rootVarDict) == 0:
    rootVarDict.append(VarDict.VarDict())
  return rootVarDict[0]


rootVarDict = list()
def GetBuildVars():
  global rootVarDict

  if len(rootVarDict) == 0:
    rootVarDict.append(VarDict.VarDict())
  return rootVarDict[-1]

def CheckpointBuildVars():
  global rootVarDict
  import copy
  newBuildVars = copy.copy(GetBuildVars())
  rootVarDict.append(newBuildVars)
  logging.debug("Created checkpoint {0} for build vars".format(len(rootVarDict)))

def RevertBuildVars():
  global rootVarDict
  rootVarDict.pop()
  logging.debug("Reverting to checkpoint {0} for build vars".format(len(rootVarDict)))

def ClearEnvironment():
  global rootEnvironment
  rootEnvironment = None



## @file EnvironmentDescriptorFiles.py
# This module contains code for working with the JSON environment
# descriptor files. It can parse the files, validate them, and return
# objects representing their contents.
#
##
# Copyright (c) 2017, Microsoft Corporation
#
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##
import os
import yaml
import sys


class PathEnv(object):
  def __init__(self, descriptor):
    super(PathEnv, self).__init__()

    #
    # Set the data for this object.
    #
    self.scope = descriptor['scope']
    self.flags = descriptor['flags']
    self.var_name = descriptor.get('var_name', None)

    self.descriptor_location = os.path.dirname(descriptor['descriptor_file'])
    self.published_path = self.descriptor_location


class DescriptorFile(object):
  def __init__(self, file_path):
    super(DescriptorFile, self).__init__()

    self.file_path = file_path
    self.descriptor_contents = None

    with open(file_path, 'r') as file:
      try:
        self.descriptor_contents = yaml.safe_load(file)
      except:
        pass  # We'll pick up this error when looking at the data.

    #
    # Make sure that we loaded the file successfully.
    #
    if self.descriptor_contents is None:
      raise ValueError("Could not load contents of descriptor file '%s'!" % file_path)

    # The file path is an implicit descriptor field.
    self.descriptor_contents['descriptor_file'] = self.file_path

    # All files require a scope.
    if 'scope' not in self.descriptor_contents:
      raise ValueError("File '%s' missing required field '%s'!" % (self.file_path, 'scope'))

    # If a file has flags, make sure they're sane.
    if 'flags' in self.descriptor_contents:
      # If a flag requires a name, make sure a name is provided.
      for name_required in ('set_shell_var', 'set_build_var'):
        if name_required in self.descriptor_contents['flags']:
          if 'var_name' not in self.descriptor_contents:
            raise ValueError("File '%s' has a flag requesting a var, but does not provide 'var_name'!" % self.file_path)
    
    # clean up each string item for more reliable processing
    for (k,v) in self.descriptor_contents.items():
      if(self.is_string(v)):
        self.descriptor_contents[k] = self.sanitize_string(v)
    
  #
  # Check if input is a string type
  #
  def is_string(self, s):
    # if we use Python 3
    if (sys.version_info[0] >= 3):
        return isinstance(s, str)
    # we use Python 2
    return isinstance(s, basestring)

  #
  # Clean up a string "value" in the descriptor file.
  #
  def sanitize_string(self, s):
    # Perform any actions needed to clean the string.
    return s.strip()


class PathEnvDescriptor(DescriptorFile):
  def __init__(self, file_path):
    super(PathEnvDescriptor, self).__init__(file_path)

    #
    # Validate file contents.
    #
    # Make sure that the required fields are present.
    for required_field in ('flags',):
      if required_field not in self.descriptor_contents:
        raise ValueError("File '%s' missing required field '%s'!" % (self.file_path, required_field))


class ExternDepDescriptor(DescriptorFile):
  def __init__(self, file_path):
    super(ExternDepDescriptor, self).__init__(file_path)

    #
    # Validate file contents.
    #
    # Make sure that the required fields are present.
    for required_field in ('scope', 'type', 'name', 'source', 'version'):
      if required_field not in self.descriptor_contents:
        raise ValueError("File '%s' missing required field '%s'!" % (self.file_path, required_field))

class PluginDescriptor(DescriptorFile):
  def __init__(self, file_path):
    super(PluginDescriptor, self).__init__(file_path)

    #
    # Validate file contents.
    #
    # Make sure that the required fields are present.
    for required_field in ('scope', 'name', 'module'):
      if required_field not in self.descriptor_contents:
        raise ValueError("File '%s' missing required field '%s'!" % (self.file_path, required_field))
      
    # Make sure the module item doesn't have .py on the end
    if(self.descriptor_contents["module"].lower().endswith(".py")):
        self.descriptor_contents["module"] = self.descriptor_contents["module"][:-3]  #remove last 3 chars




## @file ExternalDependencies.py
# This module contains helper objects that can manipulate,
# retrieve, validate, and clean external dependencies for the
# build envrionment.
#
##
# Copyright (c) 2017, Microsoft Corporation
#
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##
import os
import logging
from PythonLibrary.UtilityFunctions import RunCmd
import shutil
import yaml
import time
from io import StringIO


class ExternalDependency(object):
  def __init__(self, descriptor):
    super(ExternalDependency, self).__init__()

    #
    # Set the data for this object.
    #
    self.scope = descriptor['scope']
    self.type = descriptor['type']
    self.name = descriptor['name']
    self.source = descriptor['source']
    self.version = descriptor['version']
    self.flags = descriptor.get('flags', None)
    self.var_name = descriptor.get('var_name', None)

    self.descriptor_location = os.path.dirname(descriptor['descriptor_file'])
    self.contents_dir = os.path.join(self.descriptor_location, self.name + "_extdep")
    self.state_file_path = os.path.join(self.contents_dir, "extdep_state.json")
    self.published_path = self.contents_dir

    if self.flags and "include_separator" in self.flags:
      self.published_path += os.path.sep

  def _clean_directory(self, dir_path):
    retry = 1
    while True:
      try:
        shutil.rmtree(dir_path)
      except OSError:
        if not retry:
          # If we're out of retries, bail.
          raise
        time.sleep(5)
        retry -= 1
        continue
      break

  def clean(self):
    logging.debug("Cleaning dependency directory for '%s'..." % self.name)
    if os.path.isdir(self.contents_dir):
      self._clean_directory(self.contents_dir)

  def fetch(self):
    # The base class does not implement a fetch.
    logging.critical("Fetch() CALLED ON BASE EXTDEP CLASS!")
    pass

  def verify(self):
    # The base class does not implement a verify.
    logging.critical("Verify() CALLED ON BASE EXTDEP CLASS!")
    return False

  def update_state_file(self):
    with open(self.state_file_path, 'w+') as file:
      yaml.dump({'version': self.version}, file)


class NugetDependency(ExternalDependency):
  global_cache_path = None

  @staticmethod
  def normalize_version(version):
    version_parts = tuple(int(num) for num in version.split('.'))
    if len(version_parts) > 4:
      raise RuntimeError("Unparsable version '%s'!")

    # Remove extra trailing zeros (beyond 3 elements).
    if len(version_parts) == 4 and version_parts[3] == 0:
      version_parts = version_parts[0:2]

    # Add missing trailing zeros (below 3 elements).
    if len(version_parts) < 3:
      version_parts = version_parts + (0,) * (3 - len(version_parts))

    # Return reformed version.
    return ".".join((str(num) for num in version_parts))

  def _fetch_from_cache(self, package_name):
    result = False
    NugetPath = os.path.join(os.environ["NUGET_PATH"], "nuget.exe")

    #
    # We still need to use Nuget to figure out where the
    # "global-packages" cache is on this machine.
    #
    if NugetDependency.global_cache_path is None:
      if os.name == "posix":
        cmd = ["mono"]
      else:
        cmd = []
      cmd += [NugetPath, "locals", "global-packages", "-list"]
      cmd_string = " ".join(cmd)
      return_buffer = StringIO()
      if (RunCmd(cmd_string, outstream=return_buffer) == 0):
        # Seek to the beginning of the output buffer and capture the output.
        return_buffer.seek(0)
        return_string = return_buffer.read()
        NugetDependency.global_cache_path = return_string.strip().strip("global-packages: ")

    #
    # If the path couldn't be found, we can't do anything else.
    #
    if not os.path.isdir(NugetDependency.global_cache_path):
      logging.info("Could not determine Nuget global packages cache location.")
      return False

    #
    # Now, try to locate our actual cache path
    nuget_version = NugetDependency.normalize_version(self.version)
    cache_search_path = os.path.join(NugetDependency.global_cache_path, package_name.lower(), nuget_version, package_name)
    if os.path.isdir(cache_search_path):
      logging.info("Local Cache found for Nuget package '%s'. Skipping fetch." % package_name)
      shutil.copytree(cache_search_path, self.contents_dir)
      self.update_state_file()
      result = True

    return result

  def fetch(self):
    package_name = self.name

    #
    # Before trying anything with Nuget feeds,
    # check to see whether the package is already in
    # our local cache. If it is, we avoid a lot of
    # time and network cost by copying it directly.
    #
    if self._fetch_from_cache(package_name):
      # We successfully found the package in the cache.
      # Bail.
      return

    #
    # If we are still here, the package wasn't in the cache.
    # We need to ask Nuget to find it.
    #

    #
    # First, fetch the contents of the package.
    #
    NugetPath = os.path.join(os.environ["NUGET_PATH"], "nuget.exe")
    temp_directory = self.get_temp_dir()
    if os.name == "posix":
      cmd = ["mono"]
    else:
      cmd = []
    cmd += [NugetPath, "install", package_name]
    cmd += ["-Source", self.source]
    cmd += ["-ExcludeVersion"]
    cmd += ["-Version", self.version]
    # cmd += ["-DirectDownload", "-NoCache"]  #avoid cache -- not sure this is good
    #cmd += ["-NonInteractive"]
    cmd += ["-Verbosity", "detailed"]
    cmd += ["-OutputDirectory", '"'+temp_directory+'"']
    cmd_string = " ".join(cmd)
    RunCmd(cmd_string)

    #
    # Next, copy the contents of the package to the
    # final resting place.
    #
    # Depending on packaging, the package content will be in one of two
    # possible locations:
    # 1. temp_directory\package_name\package_name\
    # 2. temp_directory\package_name\
    #
    source_dir = os.path.join(temp_directory, package_name, package_name)
    if not os.path.isdir(source_dir):
      source_dir = os.path.join(temp_directory, package_name)
    shutil.move(source_dir, self.contents_dir)

    #
    # Add a file to track the state of the dependency.
    #
    self.update_state_file()

    #
    # Finally, delete the temp directory.
    #
    self._clean_directory(temp_directory)

  def get_temp_dir(self):
    return self.contents_dir + "_temp"

  def clean(self):
    super(NugetDependency, self).clean()
    if os.path.isdir(self.get_temp_dir()):
      self._clean_directory(self.get_temp_dir())

  def verify(self):
    result = True
    state_data = None

    # See whether or not the state file exists.
    if not os.path.isfile(self.state_file_path):
      result = False

    # Attempt to load the state file.
    if result:
      with open(self.state_file_path, 'r') as file:
        try:
          state_data = yaml.safe_load(file)
        except:
          pass
    if state_data is None:
      result = False

    # If loaded, check the version.
    if result and state_data['version'] != self.version:
      result = False

    logging.debug("Verify '%s' returning '%s'." % (self.name, result))
    return result


def ExtDepFactory(descriptor):
  if descriptor['type'] == 'nuget':
    return NugetDependency(descriptor)
  else:
    raise ValueError("Unknown extdep type '%s' requested!" % descriptor['type'])


## @file VarDict.py
# This module contains code for a special overridable dictionary.  
# This stores most of the build configuration data and allows 
# extensive config sharing for the build process, pre-build, and 
# post-build.  
#
##
# Copyright (c) 2018, Microsoft Corporation
#
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##
from __future__ import print_function
import logging


class EnvEntry(object):
    def __init__(self, value, comment, overridable=False):
        self.Value = value
        self.Comment = comment
        self.Overrideable = overridable

    def PrintEntry(self, f=None):
        if(f == None):
            print("Value: %s" % self.Value, file=f)
            print("Comment: %s" % self.Comment, file=f)
            if(self.Overrideable):
                print("Value overridable", file=f)
            print("**********************", file=f)
    #
    # Function used to override the value if option allows it
    #
    def SetValue(self, value, comment, overridable = False):
        if (value == self.Value):
            return True

        if(not self.Overrideable):
            logging.debug("Can't set value [%s] as it isn't overrideable. Previous comment %s" % (value,self.Comment))
            return False

        self.Value = value
        self.Comment = comment
        self.Overrideable = overridable
        return True

    def GetValue(self):
        return self.Value


class VarDict(object):
    def __init__(self):
        self.Logger = logging.getLogger("EnvDict")
        self.Dstore = {}     #a set of envs

    def GetEntry(self, key):
        return self.Dstore.get(key.upper())

    def __copy__(self):
        new_copy = VarDict()
        new_copy.Logger = self.Logger

        new_copy.Dstore = {}
        for key in self.Dstore:
            entry = self.GetEntry(key)
            value = entry.Value
            comment = entry.Comment
            override = entry.Overrideable
            new_copy.SetValue(key,value,comment,override)
        return new_copy


    def GetValue(self, k):
        key = k.upper()
        en = self.GetEntry(key)
        if(en != None):
            self.Logger.debug("Key %s found.  Value %s" % (key, en.GetValue()))
            return en.GetValue()
        else:
            self.Logger.debug("Key %s not found" % key)
            return None

    def SetValue(self, k, v, comment, overridable=False):
        key = k.upper()
        en = self.GetEntry(key)
        value = str(v)
        self.Logger.debug("Trying to set key %s to value %s" % (k, v))
        if(en == None):
            #new entry
            en = EnvEntry(value, comment, overridable)
            self.Dstore[key] = en
            return True
        
        return en.SetValue(value, comment, overridable)


    #
    # function used to get a build var value for given key and buildtype
    #
    # if BuildType is None 
    # Build vars are defined by vars that start with BLD_
    #  BLD_*_<YOUR KEY HERE> means all build types
    #  BLD_DEBUG_<YOUR KEY HERE> means build of debug type
    #  BLD_RELEASE_<YOUR KEY HERE> means build of release type
    #  etc
    #
    def GetBuildValue(self, key, BuildType=None):
        rv = None
        
        if(BuildType == None):
            BuildType = self.GetValue("TARGET")
        
        if(BuildType == None):
            logging.debug("GetBuildValue - Invalid Parameter BuildType is None and Target Not set. Key is: " + key)
            return None

        if(key == None):
            logging.debug("GetBuildValue - Invalid Parameter key is None. BuildType is: " + BuildType)
            return None

        
        ty = BuildType.upper().strip()
        tk = key.upper().strip()
        # see if specific 
        k = "BLD_" + ty + "_" + tk
        rv = self.GetValue(k)
        if(rv == None):
            #didn't fine build type specific so check for generic
            k = "BLD_*_" + tk
            rv = self.GetValue(k)

        #return value...if not found should return None
        return rv

    #
    # function used to get a dictionary for all build vars
    #
    # Build vars are defined by vars that start with BLD_
    #  BLD_*_<YOUR VAR HERE> means all build types
    #  BLD_DEBUG_<YOUR VAR HERE> means build of debug type
    #  BLD_RELEASE_<YOUR VAR HERE> means build of release type
    #  etc
    #
    def GetAllBuildKeyValues(self, BuildType=None):
        l = {}
        if(BuildType == None):
            BuildType = self.GetValue("TARGET")

        if(BuildType == None):
            logging.debug("GetAllBuildKeyValues - Invalid Parameter BuildType is None and Target Not Set.")
            return l

        ty = BuildType.upper().strip()
        logging.debug("Getting all build keys for build type " + ty)

        #get all the generic build options
        for key,value in self.Dstore.items():
            if(key.startswith("BLD_*_")):
                k = key[6:]
                l[k] = value.GetValue()
                
        #will override with specific for this build type
        #figure out offset part of key name to strip
        ks = len(ty) + 5
        for key,value in self.Dstore.items():
            if(key.startswith("BLD_" + ty + "_")):
                k = key[ks:]
                l[k] = value.GetValue()

        return l

    def PrintAll(self, fp=None):
        f = None
        if(fp is not None):
            f = open(fp, 'w')
        for key,value in self.Dstore.items():
            print("Key = %s"% key, file=f)
            value.PrintEntry(f)
        if(f):
            f.close()


## @file SelfDescribingEnvironment.py
# This module contains code that is capable of scanning the source tree for
# files that describe the source and dependencies and acting upon those files.
#
##
# Copyright (c) 2017, Microsoft Corporation
#
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##
import os, sys
from MuEnvironment import ShellEnvironment
from MuEnvironment import EnvironmentDescriptorFiles as EDF
from MuEnvironment import ExternalDependencies
import logging

ENVIRONMENT_BOOTSTRAP_COMPLETE = False
ENV_STATE = None

class SelfDescribingEnvironment(object):
  def __init__(self, workspace_path, scopes=()):
    super(SelfDescribingEnvironment, self).__init__()

    self.workspace = workspace_path

    # Determine the final set of scopes.
    # Start with the provided set.
    self.scopes = scopes
    # Add any OS-specific scope.
    if os.name == 'nt':
      self.scopes += ('global-win',)
    elif os.name == 'posix':
      self.scopes += ('global-nix',)
    # Add the global scope.
    self.scopes += ('global',)
    # Validate that all scopes are unique.
    if len(self.scopes) != len(set(self.scopes)):
      raise ValueError("All scopes must be unique!")

    self.paths = None
    self.extdeps = None
    self.plugins = None

  def _gather_env_files(self, ext_strings, base_path):
    # Make sure that the search extension matches easily.
    search_files = tuple(ext_string.lower() for ext_string in ext_strings)

    # Walk all of the directories under base_path and find all files
    # matching the extension.
    matches = {}
    for root, dirs, files in os.walk(base_path, topdown=True):
      # Check to see whether any of these directories should be skipped.
      # TODO: Allow these to be passed in via arguments.
      for index, dir in enumerate(dirs):
        if dir == '.git':
          del dirs[index]

      # Check for any files that match the extensions we're looking for.
      for file in files:
        for search_file in search_files:
          if file.lower().endswith(search_file+".json") or file.lower().endswith(search_file+".yaml"):
            if search_file in matches:
              matches[search_file].append(os.path.join(root, file))
            else:
              matches[search_file] = [os.path.join(root, file)]

    return matches

  def load_workspace(self):
    logging.debug("--- SelfDescribingEnvironment.load_workspace()")
    logging.debug("Loading workspace: %s" % self.workspace)
    logging.debug("  Including scopes: %s" % ', '.join(self.scopes))

    #
    # First, we need to get all of the files that describe our environment.
    #
    env_files = self._gather_env_files(('path_env', 'ext_dep', 'plug_in'), self.workspace)

    #
    # Now that the files have been found, load them, sort them, and filter them
    # so they can be applied to the environment.
    #
    def _sort_and_filter_descriptors(class_type, file_list, scopes):
      all_descriptors = tuple(class_type(desc_file).descriptor_contents for desc_file in file_list)

      known_ids = {}
      active_overrides = {}
      final_list = []

      for scope in scopes:
        for descriptor in all_descriptors:
          # If this descriptor isn't in the current scope, we can ignore it for now.
          if descriptor['scope'].lower() != scope.lower():
            continue

          cur_file = descriptor['descriptor_file']

          # If this descriptor has an ID, we need to check for overrides and collisions.
          if 'id' in descriptor:
            cur_id = descriptor['id'].lower()

            # First, check for overrides. There's no reason to process this file if it's being overridden.
            if cur_id in active_overrides:
              logging.debug("Descriptor '%s' is being overridden by descriptor '%s' based on ID '%s'." % (cur_file, active_overrides[cur_id], cur_id))
              continue

            # Next, check for ID collisions.
            if cur_id in known_ids:
              raise RuntimeError(
                "Descriptor '%s' shares the same ID '%s' with descriptor '%s'." % (cur_file, cur_id, known_ids[cur_id])
                )

            # Finally, we can add this file to the known IDs list.
            known_ids[cur_id] = cur_file

          # If we're still processing, we can add this descriptor to the output.
          logging.debug("Adding descriptor '%s' to the environment with scope '%s'." % (cur_file, scope))
          final_list.append(descriptor)

          # Finally, check to see whether this descriptor overrides anything else.
          if 'override_id' in descriptor:
            cur_override_id = descriptor['override_id'].lower()
            # If we're attempting to override someting that's already been processed,
            # we should spit out a warning of sort.
            if cur_override_id in known_ids:
              logging.warning("Descriptor '%s' is trying to override iID '%s', but it's already been processed." % (cur_file, cur_override_id))
            active_overrides[cur_override_id] = descriptor['descriptor_file']

      return tuple(final_list)

    if 'path_env' in env_files:
      self.paths = _sort_and_filter_descriptors(EDF.PathEnvDescriptor, env_files['path_env'], self.scopes)

    if 'ext_dep' in env_files:
      self.extdeps = _sort_and_filter_descriptors(EDF.ExternDepDescriptor, env_files['ext_dep'], self.scopes)

    if 'plug_in' in env_files:
      self.plugins = _sort_and_filter_descriptors(EDF.PluginDescriptor, env_files['plug_in'], self.scopes)

    return self

  # This is a generator to reduce code duplication when wrapping the pathenv objects.
  def _get_paths(self):
    if self.paths is not None:
      # Apply in reverse order to get the expected hierarchy.
      for path_descriptor in reversed(self.paths):
        # Use the helper factory to get an object
        # capable of managing each dependency.
        yield EDF.PathEnv(path_descriptor)

  # This is a generator to reduce code duplication when wrapping the extdep objects.
  def _get_extdeps(self):
    if self.extdeps is not None:
      # Apply in reverse order to get the expected hierarchy.
      for extdep_descriptor in reversed(self.extdeps):
        # Use the helper factory to get an object
        # capable of managing each dependency.
        yield ExternalDependencies.ExtDepFactory(extdep_descriptor)

  def _apply_descriptor_object_to_env(self, desc_object, env_object):
    # Walk through each possible environment modification
    # and apply to the environment as required.
    if 'set_path' in desc_object.flags:
      env_object.insert_path(desc_object.published_path)
    if 'set_pypath' in desc_object.flags:
      env_object.insert_pypath(desc_object.published_path)
    if 'set_build_var' in desc_object.flags:
      env_object.set_build_var(desc_object.var_name, desc_object.published_path)
    if 'set_shell_var' in desc_object.flags:
      env_object.set_shell_var(desc_object.var_name, desc_object.published_path)

  def update_simple_paths(self, env_object):
    logging.debug("--- SelfDescribingEnvironment.update_simple_paths()")
    for path in self._get_paths():
      self._apply_descriptor_object_to_env(path, env_object)

  def update_extdep_paths(self, env_object):
    logging.debug("--- SelfDescribingEnvironment.update_extdep_paths()")
    for extdep in self._get_extdeps():
      self._apply_descriptor_object_to_env(extdep, env_object)

  def update_extdeps(self, env_object):
    logging.debug("--- SelfDescribingEnvironment.update_extdeps()")
    for extdep in self._get_extdeps():
      # Check to see whether it's necessary to fetch the files.
      if not extdep.verify():
        extdep.clean()
        extdep.fetch()

  def clean_extdeps(self, env_object):
    for extdep in self._get_extdeps():
      extdep.clean()
      # TODO: Determine whether we want to update the env.

  def verify_extdeps(self, env_object):
    result = True
    for extdep in self._get_extdeps():
      if not extdep.verify():
        result = False
        logging.info("Dependency '%s' is not met!" % extdep.name)

    return result


def BootstrapEnvironment(workspace, scopes=()):
  global ENVIRONMENT_BOOTSTRAP_COMPLETE, ENV_STATE

  if not ENVIRONMENT_BOOTSTRAP_COMPLETE:
    #
    # ENVIRONMENT BOOTSTRAP STAGE 1
    # Locate and load all environment description files.
    #
    build_env = SelfDescribingEnvironment(workspace, scopes).load_workspace()

    #
    # ENVIRONMENT BOOTSTRAP STAGE 2
    # Parse all of the PATH-related descriptor files to make sure that
    # any required tools or Python modules are now available.
    #
    shell_env = ShellEnvironment.GetEnvironment()
    build_env.update_simple_paths(shell_env)

    #
    # ENVIRONMENT BOOTSTRAP STAGE 3
    # Now that the preliminary paths have been loaded,
    # we can load the modules that had greater dependencies.
    #
    build_env.update_extdep_paths(shell_env)

    # Bind our current execution environment into the shell vars.
    shell_env.set_shell_var("PYTHON3", sys.executable)

    # Debug the environment that was produced.
    shell_env.log_environment()

    ENVIRONMENT_BOOTSTRAP_COMPLETE = True
    ENV_STATE = (build_env, shell_env)

  # Return the environment as it's configured.
  return ENV_STATE


def CleanEnvironment(workspace, scopes=()):
  # Bootstrap the environment.
  (build_env, shell_env) = BootstrapEnvironment(workspace, scopes)

  # Clean all the dependencies.
  build_env.clean_extdeps(shell_env)


def UpdateDependencies(workspace, scopes=()):
  # Bootstrap the environment.
  (build_env, shell_env) = BootstrapEnvironment(workspace, scopes)

  # Clean all the dependencies.
  build_env.update_extdeps(shell_env)


def VerifyEnvironment(workspace, scopes=()):
  # Bootstrap the environment.
  (build_env, shell_env) = BootstrapEnvironment(workspace, scopes)

  # Clean all the dependencies.
  return build_env.verify_extdeps(shell_env)


if __name__ == "__main__":
  # For testing, make some assumptions about where this file is still located.
  base_path = os.path.dirname(os.path.abspath(__file__))
  base_path = os.path.dirname(base_path)
  base_path = os.path.dirname(base_path)
  base_path = os.path.dirname(base_path)

  # For testing, pick a project to use as the scope.
  test_scope = ('ivanhoe',)

  # Clean the environment.
  CleanEnvironment(base_path, test_scope)

  # Update the environment.
  UpdateDependencies(base_path, test_scope)

  # Verify the environment.
  VerifyEnvironment(base_path, test_scope)
